import 'dart:convert';
import 'dart:typed_data';

import 'package:boost/boost.dart';
import 'package:cl_datahub_common/common.dart';

T? decodeTyped<T>(dynamic raw, {DTOFactory? factory}) {
  if (raw == null) {
    return null;
  }

  if (factory != null) {
    final result = factory(raw);
    if (result is! T) {
      throw ApiError('Factory returned wrong type: $result (should be $T)');
    }

    return result as T;
  }

  if (raw is T) {
    return raw;
  }

  if (T == String) {
    return raw.toString() as T;
  }

  if (T == int) {
    return int.tryParse(raw.toString()) as T;
  }

  if (T == double) {
    return double.tryParse(raw.toString()) as T;
  }

  if (T == bool) {
    if (raw is num) {
      return raw > 0 as T;
    }

    return (raw.toString().toLowerCase() == 'true') as T;
  }

  if (T == DateTime) {
    return DateTime.tryParse(raw.toString()) as T;
  }

  if (T == Uint8List) {
    return Base64Decoder().convert(raw.toString()) as T;
  }

  throw ApiError.invalidType(T);
}

dynamic encodeTyped<T>(T value) {
  if (value == null) {
    return null;
  }

  if (T == DateTime) {
    return (value as DateTime).toIso8601String();
  }

  if (T == Uint8List) {
    return Base64Encoder().convert(value as Uint8List);
  }

  if (T == String || T == int || T == double || T == bool) {
    return value;
  }

  throw ApiError.invalidType(T);
}

String buildQueryString(Map<String, String?> query) {
  if (query.isEmpty) {
    return '';
  }

  String encodeEntry(MapEntry<String, String?> e) {
    final key = Uri.encodeQueryComponent(e.key);
    if (!nullOrEmpty(e.value)) {
      return '$key=${Uri.encodeQueryComponent(e.value!)}';
    } else {
      return key;
    }
  }

  return '?' + query.entries.map(encodeEntry).join('&');
}

final RegExp _plExp = RegExp('(?<!\\\\)\\{([\\w-]*)}');

String replacePlaceholders(String source, Map<String, dynamic> values) {
//TODO maybe add encode method for values, not just toString
  return source.replaceAllMapped(_plExp, (match) {
    final key = match.group(1);
    if (key == null) {
      throw ApiException('Invalid placeholder-key: ${match.group(0)}');
    }

    if (values[key] == null) {
      throw ApiException('Missing value in url params: $key');
    }

    return Uri.encodeComponent(values[key].toString());
  });
}
